# coding=utf-8
from nekbot.utils.strings import long_message

__author__ = 'nekmo'

class User(object):

    def __init__(self, protocol, username, id=None, groupchat=None):
        self.protocol, self.username, self.groupchat = protocol, username, groupchat
        if not hasattr(self, 'id') and id is not None:
            # id puede ser un método. No queremos entonces sobrescribirlo.
            self.id = id

    def send_message(self, body, notice=False):
        raise NotImplementedError("This protocol can't send messages to users.")

    def send_warning(self, body):
        self.send_message('Warning: %s' % body, not long_message(body))

    def send_error(self, body):
        self.send_message('Error: %s' % body, not long_message(body))

    def get_id(self):
        return self.id

    def __repr__(self):
        return '<%s %s id:%s>' % (self.__class__, self.username, self.id)

    def __str__(self):
        return self.username

    def __eq__(self, other):
        return self.id == other.id

class Users(dict):
    def __init__(self, protocol):
        self.protocol = protocol
        dict.__init__(self)


# coding=utf-8
from logging import getLogger
import threading
from nekbot.utils.modules import get_module, get_main_class

__author__ = 'nekmo'

logger = getLogger('nekbot.core.modular')

class Modular(object):
    """Modular es una clase genérica para cargar módulos desde un path e instanciarlos.
    Éste se usa para Protocols y Plugins. El module_path es un string que debe tener un
    '%s'  con el path de donde se obtendrán los módulos. Por ejemplo, 'modules.%s'. Se debe
    proveer un listado con los módulos del directorio que se cargarán (modules_names) y
    éstos serán puestos en self.instances. Intentará buscarse en los módulos una "clase
    principal", con el mismo nombre del módulo, pero con camelcase, y se instanciará en
    self.instances.
    """
    ModularError = Exception
    module_path = None
    fail_on_not_instance = True

    def __init__(self, nekbot, modules_names):
        self.modules = {}  # Módulos, no las instancias.
        self.instances = {}  # Instacia ya iniciada dentro del módulo. Ej. "Telegram"
        self.nekbot = nekbot
        self.modules_names = modules_names
        if self.module_path is None:
            raise self.ModularError('Please, provide a module_path for %s' % self.__class__)

    def get_module(self, module_path):
        return get_module(module_path)

    def start(self, module_name):
        # Obtengo el módulo por el module_name
        module = self.get_module(self.module_path % module_name)
        self.modules[module_name] = module
        instance = None
        try:
            # Busco dentro de dicho módulo la "clase principal", aquella con el mismo nombre del
            # módulo, pero que comienza por mayúscula
            instance = get_main_class(module, module_name)
        except ImportError:
            if self.fail_on_not_instance:
                raise self.ModularError('Module %s has does not have a valid instance' % module_name)
        if instance is not None:
            if not hasattr(self, module_name):
                setattr(self, module_name, instance)
            try:
                instance = instance(self.nekbot)
            except Exception as e:
                logger.error('Instance for "%s" is invalid: %s' % (module_name, e))
            else:
                instance.start()  # Iniciar la instancia
                self.instances[module_name] = instance

    def close(self):
        for instance in self.instances.values():
            if hasattr(instance, 'close'):
                l = threading.Thread(target=instance.close)
                l.start()

    def start_all(self):
        for module_name in self.modules_names:
            self.start(module_name)



__author__ = 'nekmo'

def human_join(list, and_='and', sep=', '):
    if len(list) == 1:
        return list[0]
    return '%s %s %s' % (sep.join(list[:-1]), and_, list[-1])

def long_message(message, newlines=1, length=140):
    if not isinstance(message, (str, unicode)):
        message = str(message)
    if len(message) > length:
        return True
    if len(message.split('\n')) - 1 > newlines:
        return True
    return False


# coding=utf-8
from Queue import Queue
from collections import defaultdict
import re
import threading
import datetime
from nekbot.core.exceptions import PrintableException

__author__ = 'nekmo'

class CancelTemp: pass

class TempTimeout(PrintableException):
    pass

class TempRegexList(defaultdict):
    def __init__(self):
        defaultdict.__init__(self, list)
        self.delete_timer = None
        # comandos ordenados por tiempo para ser eliminados en su momento
        self.to_delete = defaultdict(list)

    def start_delete_timer(self):
        if self.delete_timer is not None:
            self.delete_timer.cancel()
        timers = sorted(self.to_delete.keys())
        if not timers: return
        timer = timers[0]
        t = (timer - datetime.datetime.now()).seconds
        if t < 0: t = 0
        self.delete_timer = threading.Timer(t, self.start_deletion, (timer,))
        self.delete_timer.start()

    def start_deletion(self, timer):
        temp_regex_list = self.to_delete[timer]
        for temp_regex in temp_regex_list:
            self.delete(temp_regex, TempTimeout('Timeout: exceeded %i seconds' % temp_regex.timeout))
        if timer in self:
            del self.to_delete[timer]
        self.start_delete_timer()

    def register(self, temp_regex):
        # Registrando por su tiempo para eliminación para borrarlo si supera el tiempo
        self.to_delete[temp_regex.delete_t].append(temp_regex)
        # Registrando por su pattern para poder hacer el match fácilmente
        self[temp_regex.pattern].append(temp_regex)
        # Se reinicia el tiempo para el próximo borrado
        self.start_delete_timer()

    def delete(self, temp_regex, msg=None):
        if msg is None: msg = CancelTemp
        temp_regex.write(msg)
        t = temp_regex.delete_t
        reload_timer = False
        if t in self.to_delete and temp_regex in self.to_delete[t]:
            self.to_delete[t].remove(temp_regex)
            if not self.to_delete[t]:
                # No hay más para esta hora. Se borra el listado.
                del self.to_delete[t]
                reload_timer = True
        if temp_regex.pattern in self and temp_regex in self[temp_regex.pattern]:
            self[temp_regex.pattern].remove(temp_regex)
            if not self[temp_regex.pattern]:
                # No hay más para esta hora. Se borra el listado.
                del self[temp_regex.pattern]
                reload_timer = True
            if not self.to_delete[temp_regex.delete_t]:
                # No hay más para esta hora. Se borra el listado.
                del self.to_delete[temp_regex.delete_t]
                reload_timer = True
        if reload_timer:
            # Ha habido un cambio en la línea cronológica. Volver a iniciar timer
            self.start_delete_timer()

    def match_message(self, msg):
        for pattern, temp_regex_list in self.items():
            if not pattern.match(msg.body): continue
            match = re.findall(pattern, msg.body)
            for temp_regex in temp_regex_list:
                new_msg = msg.copy()
                new_msg.match = match
                temp_regex.write(new_msg)

temp_regex_patterns = TempRegexList()

class TempRegex(object):
    def __init__(self, protocol, pattern, user=None, timeout=300, no_raise=False):
        if isinstance(pattern, (str, unicode)):
            pattern = re.compile(pattern)
        self.no_raise = no_raise
        self.pattern = pattern
        self.protocol = protocol
        self.user = user
        self.timeout = timeout
        self.delete_t = datetime.datetime.now() + datetime.timedelta(seconds=timeout)
        self.queue = Queue()
        # Registrar este nuevo TempRegex
        temp_regex_patterns.register(self)

    def write(self, msg):
        self.queue.put(msg)

    def read(self):
        while self.delete_t > datetime.datetime.now():
            msg = self.queue.get()
            if isinstance(msg, CancelTemp):
                break
            if isinstance(msg, Exception):
                if self.no_raise:
                    yield msg
                else:
                    raise msg
            yield msg
        yield CancelTemp

    def done(self):
        temp_regex_patterns.delete(self)


# coding=utf-8
from logging import getLogger
import logging
from nekbot import settings

from nekbot.core.modular import Modular
from .base.message import Message
from .base.user import User
from .base.group_chat import GroupChat
from .base import Protocol
from nekbot.utils.modules import get_module

if '__path__' in globals():
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)

__author__ = 'nekmo'

logger = getLogger('nekbot.protocols')

class Protocols(Modular):
    module_path = 'nekbot.protocols.%s'

    def __init__(self, nekbot, protocols):
        Modular.__init__(self, nekbot, protocols)

    def load_settings(self, protocol):
        try:
            module_settings = get_module('nekbot.protocols.%s.global_settings' % protocol)
        except ImportError:
            logger.debug('%s protocol does not have settings' % protocol)
        else:
            logger.debug('Loaded %s protocol settings' % protocol)
            settings.write_conf(module_settings, True)

    def start(self, protocol):
        self.load_settings(protocol)
        logger.info('Starting %s protocol', protocol)
        Modular.start(self, protocol)


# coding=utf-8
__author__ = 'nekmo'

def get_module(path):
    try:
        return __import__(path, globals(), locals(), [path.split('.')[-1]])
    except ImportError:
        # Puede ser un método  o propiedad
        module = '.'.join(path.split('.')[:-1])
        module = __import__(module, globals(), locals(), [module.split('.')[-1]])
        return getattr(module, path.split('.')[-1])

def get_main_class(module, name):
    if hasattr(module, name.capitalize()):
        instance = getattr(module, name.capitalize())
    else:
        raise ImportError
    return instance



# coding=utf-8
import platform
from nekbot.core.commands import command

__author__ = 'nekmo'


@command
def about(msg):
    """Información sobre este bot.
    """
    body = u'⚛ Este bot funciona gracias a NekBot %s bajo Python %s. Protocolos en uso: %s' % (
        msg.protocol.nekbot.version, platform.python_version(),
        ', '.join([protocol for protocol in msg.protocol.nekbot.protocols.modules_names])
    )
    msg.reply(body)


import threading
from nekbot.core import events

__author__ = 'nekmo'


class Protocol(threading.Thread):
    features = []

    def __init__(self, nekbot):
        self.nekbot = nekbot
        threading.Thread.__init__(self)
        self.name = self.__class__.__name__.lower()
        self.daemon = True

        self.init()

    def prepare_message(self, body):
        if not isinstance(body, (str, unicode)):
            body = str(body)
        return body

    def propagate(self, event, *args, **kwargs):
        events.propagate(event, self, *args, **kwargs)

    def close(self):
        pass


# coding=utf-8
from collections import defaultdict
import threading

__author__ = 'nekmo'

class Events(defaultdict):
    def __init__(self):
        super(Events, self).__init__(list)

    def propagate(self, event, *args, **kwargs):
        if not event in self: return
        for function in self[event]:
            # TODO: Es necesario limitar el tiempo, hacer workers...
            l = threading.Thread(target=function, args=args, kwargs=kwargs)
            l.start()

events = Events()

def event(event_name):
    """Decorador para que se ejecute una función con un determinado evento.

    Uso:
        @event('mievento')
        def function():
            pass
    """
    def decorator(f):
        events[event_name].append(f)
        return f
    return decorator




class AuthAddress(object):
    def __init__(self, address):
        self._user = None
        self._host = None
        self._port = None
        self._endpoint = None
        self.address = address

    @property
    def user(self):
        if self._user is None:
            user_endpoint = self.address.split('@')
            if len(user_endpoint) > 1:
                self._user = user_endpoint[0]
            else:
                self._user = False
        return self._user

    @property
    def endpoint(self):
        if self._endpoint is None:
            self._endpoint = self.address.split('@')
            if len(self.endpoint) > 1:
                self._endpoint = self.endpoint[1]
            else:
                self._endpoint = self._endpoint[0]
        return self._endpoint

    @property
    def host(self):
        if self._host is None:
            self._host = self.endpoint.split(':')[0]
        return self._host

    @property
    def port(self):
        if self._port is None:
            host_port = self.endpoint.split(':')
            if len(host_port) < 2:
                self._port = False
            else:
                self._port = host_port[1]
        return self._port

    def __str__(self):
        return self.address

class UserPassword(object):
    def __init__(self, user_password):
        self._user = None
        self._password = None
        self.user_password = user_password

    @property
    def user(self):
        if self._user is None:
            self._user = self.user_password.split(':')
        return self._user

    @property
    def password(self):
        if self._password is None:
            user_password = self.user_password.split(':')
            if len(user_password) > 1:
                self._password = user_password[1]
            else:
                self._password = False
        return self._password



import os
from nekbot.core.exceptions import SecurityError, InvalidArgument

__author__ = 'nekmo'

class Node(object):
    def __init__(self, root):
        root = os.path.abspath(root)
        if not os.path.exists(root):
            raise ValueError('Invalid root for %s: "%s"' % (self.__class__.__name__,
                                                            root))
        self.root = root

    def path_not_exist(self, node):
        raise InvalidArgument('The path does not exist.', node)

    def additional_checks(self, node):
        return node

    def __call__(self, node):
        node = os.path.abspath(os.path.join(self.root, node))
        if not node.startswith(self.root):
            raise SecurityError('The path does not belong to the base path.', node)
        if not os.path.exists(node):
            return self.path_not_exist(node)
        return self.additional_checks(node)


class File(Node):
    def additional_checks(self, node):
        if not os.path.isfile(node):
            raise InvalidArgument('The path is not a file.', node)
        return node


class Dir(Node):
    def additional_checks(self, node):
        if not os.path.isdir(node):
            raise InvalidArgument('The path is not a directory.', node)
        return node


# coding=utf-8
from collections import defaultdict
from logging import getLogger
import shlex
import traceback
import threading
from nekbot import settings
from nekbot.core.commands.argparse import ArgParse
from nekbot.core.exceptions import PrintableException
from nekbot.utils.decorators import optional_args

__author__ = 'nekmo'

logger = getLogger('nekbot.core.commands')

def get_arguments(body):
    """Dividir una cadena de texto en un listado de argumentos como en un shell.
    """
    return shlex.split(body)

class Command(object):
    symbol = True

    def __init__(self, name=None, function=None, symbol=None, *args):
        self.name = name
        self.function = function
        self.symbol = symbol if symbol is not None else self.symbol
        self.argparse = self.get_argparse(args, function)

    def get_argparse(self, arg_types, function):
        argparse = ArgParse()
        argparse.set_arg_types(arg_types)
        argparse.set_from_function(function)
        # argparse.get_from_function(self.function)
        return argparse

    def execute(self, msg):
        if not hasattr(msg, 'args'):
            msg.args = get_arguments(msg.body)[1:]
        try:
            args = self.argparse.parse(msg.args)
        except Exception as e:
            return msg.short_reply(e)
        try:
            self.control(msg)
        except PrintableException as e:
            return msg.user.send_warning(e)
        try:
            response = self.function(msg, *args)
        except PrintableException as e:
            response = str(e)
        except Exception:
            logger.error(traceback.format_exc())
            msg.user.send_warning('El comando %s no finalizó correctamente.' % repr(self))
            return
        if response is not None:
            msg.reply(response)

    def control(self, msg):
        if hasattr(self.function, 'control'):
             return self.function.control.check(msg)
        if hasattr(self.function, 'command_decorator') and \
                hasattr(self.function.command_decorator, 'control'):
            return self.function.command_decorator.control.check(msg)
        return True

    def __repr__(self):
        if self.symbol:
            return settings.SYMBOL + self.name
        else:
            return self.name


class Commands(defaultdict):
    def __init__(self):
        super(Commands, self).__init__(list)

    def incoming(self, msg):
        if msg.is_from_me:
            return
        if not msg.body.startswith(settings.SYMBOL):
            return
        if msg.historical:
            return
        args = get_arguments(msg.body)
        if not args[0] in self:
            # No es un comando, se ignora
            return
        cmd, args = args[0], args[1:]
        msg.args = args
        for command in self[cmd]:
            l = threading.Thread(target=command.execute, args=(msg,))
            l.start()

    def add_command(self, name, function, *args, **kwargs):
        cmd = Command(name, function, *args, **kwargs)
        self[repr(cmd)].append(cmd)

cmds = Commands()


@optional_args
def command(func, *args, **kwargs):
    return func(*args, **kwargs)


@optional_args
class command:
    def __init__(self, *args, **kwargs):
        self.args = args
        self.kwargs = kwargs
        if len(args) > 1:
            name = args[1]
        else:
            name = args[0].func_name
        args[0].command_decorator = self
        self.name = name
        cmds.add_command(name, args[0], *args[1:], **kwargs)

    def __repr__(self):
        return '<@Command %s>' % self.name

    def __call__(self, func):
        if hasattr(self, 'control'): print('!!!')
        return func(*self.args, **self.kwargs)


# coding=utf-8
from logging import getLogger
from nekbot.core.modular import Modular
from nekbot.utils.modules import get_module

logger = getLogger('nekbot.plugins')

if '__path__' in globals():
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)

class Plugins(Modular):
    module_path = 'plugins.%s'
    fail_on_not_instance = False

    def __init__(self, nekbot, protocols):
        Modular.__init__(self, nekbot, protocols)

    def get_module(self, module_path):
        """Probar primero a cargar el módulo desde el directorio plugins
        del proyecto. Si no funciona, cargarlo desde NekBot.
        """
        try:
            return get_module(module_path)
        except ImportError:
            return get_module('nekbot.' + module_path)

    def start(self, protocol):
        logger.info('Loading %s plugin', protocol)
        Modular.start(self, protocol)


# coding=utf-8
from logging import getLogger
from time import sleep
import datetime
from nekbot.plugins import Plugins

from .signals import events, event
from nekbot.protocols import Protocols
from nekbot.conf import settings
from . import handlers
from .. import __version__

__author__ = 'nekmo'
logger = getLogger('nekbot')


class NekBot(object):
    version = __version__

    def __init__(self):
        self.protocols = Protocols(self, settings.PROTOCOLS)
        self.plugins = Plugins(self, settings.PLUGINS)
        self.start_datetime = datetime.datetime.now()

    def start(self):
        self.start_protocols()
        self.start_plugins()
        return self

    def start_protocols(self):
        self.protocols.start_all()

    def start_plugins(self):
        self.plugins.start_all()

    def close_protocols(self):
        self.protocols.close()

    def close_plugins(self):
        self.plugins.close()

    def loop(self):
        logger.info('Everything has started')
        while True: sleep(0.2)

    def close(self):
        logger.info('Closing...')
        logger.debug('Closing plugins...')
        self.close_plugins()
        logger.debug('Closing protocols...')
        self.close_protocols()



from nekbot.core import event, events
from nekbot.core.commands import cmds
from nekbot.core.commands.temp import temp_regex_patterns

__author__ = 'nekmo'


@event('message')
def commands_handler(protocol, msg):
    cmds.incoming(msg)

@event('message')
def message_by_protocol(protocol, msg):
    events.propagate('%s.message' % protocol.name, protocol, msg)

@event('message')
def temp_command(protocol, msg):
    temp_regex_patterns.match_message(msg)


# coding=utf-8
__author__ = 'nekmo'

from . import global_settings

class Settings(object):
    def __init__(self, defaults=global_settings):
        self.write_conf(global_settings) # Primero guardo la configuración global

    def configure(self, module):
        """Módulo (archivo de configuración) donde se encuentran los parámetros con
        la configuración del usuario. Esta sobrescribirá la global
        """
        if isinstance(module, (unicode, str)):
            from nekbot.utils.modules import get_module
            module = get_module(module)
        self.write_conf(module)

    def write_conf(self, settings_mod, prevent_override=False):
        for setting in dir(settings_mod):
            if prevent_override and hasattr(self, setting): continue
            if setting.isupper():
                setattr(self, setting, getattr(settings_mod, setting))

settings = Settings()


# coding=utf-8
import inspect
from nekbot.core.exceptions import InvalidArgument, PrintableException

__author__ = 'nekmo'

class AllTypes(object):
    pass

ERRORS = {
    ValueError: {
        int: 'El valor debe ser un número.',
        AllTypes: 'Debe ser de tipo {type}',
    },
}

class ArgParse(object):
    def __init__(self):
        self.arg_types = []
        self.kwarg_types = []

    def set_arg_types(self, arg_types):
        self.arg_types = arg_types

    def set_from_function(self, function):
        kwargs_types = inspect.getargspec(function).defaults
        if kwargs_types:
            self.kwarg_types = map(self.get_type, kwargs_types)

    def get_type(self, value):
        if hasattr(value, '__call__'):
            return value
        elif hasattr(value.__class__, '__call__'):
            return value.__class__
        else:
            return str

    def parse_arg(self, type, value, pos=None):
        try:
            if hasattr(type, '__call__'):
                value = type(value)
        except Exception as e:
            if isinstance(e, InvalidArgument):
                # ¡Es una excepción creada por nosotros. Le damos por si lo necesitase
                # información adicional y lo lanzamos tal cual
                e.give_info(value, pos)
                raise e
            err_class = e.__class__
            # No es un tipo de excepción conocida.
            if not ERRORS.get(err_class): raise e
            # Es conocida, pero no tenemos tipo para el mismo, y no hay por defecto
            if not ERRORS[err_class].get(type) and not ERRORS[err_class].get(AllTypes):
                raise e
            # No tenemos tipo para él, pero sí por defecto
            if not ERRORS[err_class].get(type) and ERRORS[err_class].get(AllTypes):
                raise InvalidArgument(ERRORS[err_class][AllTypes], value, pos)
            # ¡Hemos triunfado! Hay para excepción->tipo
            raise InvalidArgument(ERRORS[err_class][type], value, pos)
        return value

    def parse(self, args):
        if len(self.arg_types) > len(args):
            raise PrintableException('Not enough arguments for this command. Missing %i argument%s.' % \
                                      (len(self.arg_types), 's' if len(self.arg_types) > 1 else ''))
        all_types = tuple(self.arg_types) + tuple(self.kwarg_types)
        for i, arg in enumerate(args):
            if len(all_types) <= i: break
            arg = self.parse_arg(all_types[i], arg, i)
            args[i] = arg
        return args


# coding=utf-8
from nekbot.protocols.base.event import Event
from nekbot.utils.strings import long_message

__author__ = 'nekmo'

class Message(Event):
    event_name = 'message'
    historical = False

    def __init__(self, protocol, body, user, groupchat=None):
        """Un nuevo mensaje entrante
        :param protocol: Instancia del protocol
        :param body: String con el mensaje recibido.
        :param user: Instancia de User del usuario que envía el mensaje.
        :param groupchat: Instancia de GroupChat
        """
        self.protocol, self.body, self.user, self.groupchat = protocol, body, user, groupchat

    def reply(self, body, notice=False):
        if notice and 'notice' in self.protocol.features:
            self.user.send_message(body, True)
        elif self.is_groupchat and self.is_public:
            self.groupchat.send_message(body)
        else:
            self.user.send_message(body)

    def short_reply(self, body):
        self.reply(body, not long_message(body))

    @property
    def is_from_me(self):
        if self.user.id:
            property = 'id'
        else:
            property = 'username'
        if self.is_groupchat:
            return getattr(self.groupchat.bot, property) == getattr(self.user.username, property)
        else:
            return getattr(self.protocol.bot, property) == getattr(self.user, property)

    @property
    def is_groupchat(self):
        raise NotImplementedError("This protocol doesn't know if the message is from a groupchat.")

    @property
    def is_public(self):
        raise NotImplementedError("This protocol doesn't know if the message is public.")

    @property
    def is_private(self):
        raise NotImplementedError("This protocol doesn't know if the message is private.")

    def _copy(self):
        return self.__class__(self.protocol, self.body, self.user, self.groupchat)

    def copy(self):
        return self._copy()

    def __str__(self):
        return self.body


from collections import defaultdict
from nekbot.protocols.base.user import Users

__author__ = 'nekmo'

class GroupChat(object):

    def __init__(self, protocol, name):
        self.users = Users(protocol)
        self.protocol, self.name = protocol, name

    def send_message(self, body):
        raise NotImplementedError("This protocol can't send public messages.")

    @property
    def bot(self):
        raise NotImplementedError('This protocol does not know who I am')

    def __str__(self):
        return self.name

class GroupChats(dict):
    def __init__(self, protocol):
        self.protocol = protocol
        super(GroupChats, self).__init__()



# coding=utf-8
from nekbot.utils.strings import human_join

__author__ = 'nekmo'

class ProtocolError(Exception):
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return self.msg


class PrintableException(Exception):
    base_msg = ''
    def __init__(self, msg):
        self.msg = msg

    def __str__(self):
        return ('%s: %s' % (self.base_msg, self.msg)) if self.base_msg else self.msg


class InvalidArgument(PrintableException):
    def __init__(self, msg, value, pos=None):
        self.msg, self.value, self.pos = msg, value, pos

    def one_argument(self):
        body = 'El argumento '
        if self.pos is not None:
            body += 'en la posición %s ' % (self.pos + 1)
        body += 'con valor "%s", no es válido.' % self.value
        return body

    def several_arguments(self):
        positions = self.pos if self.pos else []
        body = 'Los argumentos '
        args = []
        for i, value in enumerate(self.value):
            arg = '"%s"' % value
            if len(self.pos) >= i:
                arg += ' (posición %s)' % (positions[i] + 1)
            args.append(arg)
        body += human_join(args, 'y')
        body += ' no son válidos.'
        return body

    def give_info(self, value, pos):
        self.value, self.pos = value, pos

    def __str__(self):
        if isinstance(self.value, (list, tuple)):
            body = self.several_arguments()
        else:
            body = self.one_argument()
        body = '%s %s' % (body, self.msg)
        if self.base_msg: body = '%s: %s' % (self.base_msg, body)
        return body


    def __repr__(self):
        return '<InvalidArgument "%s">' % str(self)


class InsufficientPermissions(PrintableException):
    base_msg = 'This action could not be completed due to lack of permissions'

class SecurityError(InvalidArgument):
    base_msg = 'Security Error'


#!/usr/bin/env/python
# coding=utf-8

__version__ = 'Mirai 0.1'
__author__ = 'nekmo'

from nekbot.conf import settings
from nekbot.core import NekBot
from nekbot.protocols import Protocols

if '__path__' in globals():
    from pkgutil import extend_path
    __path__ = extend_path(__path__, __name__)


#!/usr/bin/env python
import re

PROJECT = 'nekbot'
VERSION = open('VERSION').read().replace('\n', '')
MODULES = [
    'nekbot',
]
ROOT_INCLUDE = ['requirements.txt', 'VERSION', 'LICENSE.txt']
__author__ = 'nekmo'


from setuptools import setup, find_packages
from pip.req import parse_requirements

from distutils.util import convert_path
from fnmatch import fnmatchcase

import os
import sys
import glob
import uuid

try:
    long_description = open('README', 'rt').read()
except IOError:
    long_description = ''

__dir__ = os.path.abspath(os.path.dirname(__file__))


def get_url(ir):
    if hasattr(ir, 'url'): return ir.url
    if ir.link is None: return
    return ir.link.url

requirements = parse_requirements('requirements.txt', session=uuid.uuid1())
install_requires = [str(ir.req) for ir in requirements if not not get_url(ir)]

packages = find_packages(__dir__)
# Prevent include symbolic links
for package in tuple(packages):
    path = os.path.join(__dir__, package.replace('.', '/'))
    if not os.path.exists(path): continue
    if not os.path.islink(path): continue
    packages.remove(package)

# dependency_links = []
# for i, dependency in enumerate(install_requires):
#     if not re.findall('(hg|git|svn|bzr)\+', dependency): continue
#     install_requires[i] = dependency.split('#egg=')[1]
#     dependency_links.append(dependency.replace('-e ', ''))


##############################################################################
# find_package_data is an Ian Bicking creation.

# Provided as an attribute, so you can append to these instead
# of replicating them:
standard_exclude = ('*.py', '*.pyc', '*~', '.*', '*.bak', '*.swp*')
standard_exclude_directories = ('.*', 'CVS', '_darcs', './build',
                                './dist', 'EGG-INFO', '*.egg-info')


def find_package_data(where='.', package='',
                      exclude=standard_exclude,
                      exclude_directories=standard_exclude_directories,
                      only_in_packages=True,
                      show_ignored=False):
    """
    Return a dictionary suitable for use in ``package_data``
    in a distutils ``setup.py`` file.

    The dictionary looks like::

        {'package': [files]}

    Where ``files`` is a list of all the files in that package that
    don't match anything in ``exclude``.

    If ``only_in_packages`` is true, then top-level directories that
    are not packages won't be included (but directories under packages
    will).

    Directories matching any pattern in ``exclude_directories`` will
    be ignored; by default directories with leading ``.``, ``CVS``,
    and ``_darcs`` will be ignored.

    If ``show_ignored`` is true, then all the files that aren't
    included in package data are shown on stderr (for debugging
    purposes).

    Note patterns use wildcards, or can be exact paths (including
    leading ``./``), and all searching is case-insensitive.

    This function is by Ian Bicking.
    """

    out = {}
    stack = [(convert_path(where), '', package, only_in_packages)]
    while stack:
        where, prefix, package, only_in_packages = stack.pop(0)
        for name in os.listdir(where):
            fn = os.path.join(where, name)
            if os.path.isdir(fn):
                bad_name = False
                for pattern in exclude_directories:
                    if (fnmatchcase(name, pattern)
                            or fn.lower() == pattern.lower()):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, (
                                "Directory %s ignored by pattern %s"
                                % (fn, pattern))
                        break
                if bad_name:
                    continue
                if os.path.isfile(os.path.join(fn, '__init__.py')):
                    if not package:
                        new_package = name
                    else:
                        new_package = package + '.' + name
                    stack.append((fn, '', new_package, False))
                else:
                    stack.append(
                        (fn, prefix + name + '/', package, only_in_packages)
                    )
            elif package or not only_in_packages:
                # is a file
                bad_name = False
                for pattern in exclude:
                    if (fnmatchcase(name, pattern)
                            or fn.lower() == pattern.lower()):
                        bad_name = True
                        if show_ignored:
                            print >> sys.stderr, (
                                "File %s ignored by pattern %s"
                                % (fn, pattern))
                        break
                if bad_name:
                    continue
                out.setdefault(package, []).append(prefix+name)
    return out
##############################################################################

description = 'NekBot, a modular multiprotocol bot.'
package_data = {'': ROOT_INCLUDE}

for module in MODULES:
    package_data.update(find_package_data(
        module,
        package=module,
        only_in_packages=False,
    ))

setup(
    name=PROJECT,
    version=VERSION,

    description=description,
    long_description=long_description,

    author='Nekmo',
    author_email='contacto@nekmo.com',

    url='https://bitbucket.org/Nekmo/nekbot-mirai',

    classifiers=[
        'Natural Language :: Spanish',
        'Development Status :: 1 - Planning',
        'License :: OSI Approved :: MIT License',
        'Intended Audience :: Developers',
        'Environment :: Console',
        'Topic :: Communications :: Chat',
        'Topic :: Communications :: Chat :: Internet Relay Chat',
        'Topic :: Communications :: Conferencing',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.2',
        'Programming Language :: Python :: 3.3',
        'Programming Language :: Python :: 3.4',
    ],

    platforms=['linux'],

    provides=['nekbot'],
    install_requires=install_requires,

    packages=packages,
    include_package_data=True,
    # Scan the input for package information
    # to grab any data files (text, images, etc.)
    # associated with sub-packages.
    package_data=package_data,

    download_url='https://bitbucket.org/Nekmo/nekbot-mirai/get/default.tar.gz',
    keywords=['nekbot', 'bot', 'irc', 'xmpp', 'chat', 'telegram', 'jabber'],
    scripts=[os.path.join('scripts', file) for file in os.listdir('scripts')],

    # entry_points={},

    zip_safe=False,
)


# coding=utf-8
from types import IntType
from nekbot.core.commands import command
from nekbot.core.commands.control import control
from nekbot.core.commands.temp import TempRegex
from nekbot.core.exceptions import PrintableException

__author__ = 'nekmo'

@command('argnumber', int)
def argnumber(msg, number):
    assert type(number) is IntType
    msg.reply('Entregado: %i' % number)

@command
def args(msg, *args):
    msg.reply('Entregado: %s' % ', '.join(args))

@command
def kwargnumber(msg, number=10):
    assert type(number) is IntType
    return 'Entregado: %i' % number

@command
def privatemsg(msg):
    msg.user.send_message('Un mensaje privado')

@command
def publicmsg(msg):
    msg.groupchat.send_message('Un mensaje público')

@command
def test_temp_regex(msg):
    msg.reply('Elija: [Sn]')
    temp = TempRegex(msg.protocol, '^(S|s|N|n)$', timeout=5)
    response = temp.read().next()
    msg.reply('Usted ha dicho: %s' % response.match[0])

@command
@control('root')
def need_root(msg):
    return 'Hola jefe!'

@command
@control('spam')
def need_spam(msg):
    return 'Spam! Spam! Spam!!'

@command
def raise_printable_exception(msg):
    raise PrintableException("nobody's specs the spanish inquisition!")

@command
def raise_exception(msg):
    raise Exception



import inspect
import functools

__author__ = 'nekmo'

def optional_args(func):
    ''' Allow to use decorator either with arguments or not. '''

    def isFuncArg(*args, **kw):
        return len(args) == 1 and len(kw) == 0 and (
            inspect.isfunction(args[0]) or isinstance(args[0], type))

    if isinstance(func, type):
        def class_wrapper(*args, **kw):
            if isFuncArg(*args, **kw):
                return func()(*args, **kw) # create class before usage
            return func(*args, **kw)
        class_wrapper.__name__ = func.__name__
        class_wrapper.__module__ = func.__module__
        return class_wrapper

    @functools.wraps(func)
    def func_wrapper(*args, **kw):
        if isFuncArg(*args, **kw):
            return func(*args, **kw)

        def functor(userFunc):
            return func(userFunc, *args, **kw)

        return functor

    return func_wrapper


import os
import logging
import sys
import shutil

__author__ = 'nekmo'
__dir__ = os.path.abspath(os.path.dirname(__file__))

nekbot_src_dir = os.path.dirname(os.path.dirname(__file__))
conf_src_dir = os.path.join(nekbot_src_dir, 'conf')


def copytree(src, dst, symlinks=False, ignore=None):
    for item in os.listdir(src):
        s = os.path.join(src, item)
        d = os.path.join(dst, item)
        if os.path.isdir(s):
            shutil.copytree(s, d, symlinks, ignore)
        else:
            shutil.copy2(s, d)


class Management(object):
    no_configuration_required = ['createbot']

    def __init__(self, settings=None, description=None, default_level=logging.INFO):
        self.parser = self.argument_parser(settings, description, default_level)

    def argument_parser(self, settings=None, description=None, default_level=logging.INFO):
        import argparse
        if settings is None:
            settings = os.environ.get('NEKBOT_SETTINGS_MODULE', 'settings')
        if description is None:
            from __main__ import __doc__ as description
        parser = argparse.ArgumentParser(prog='nekbot', description=description)
        # Niveles de logging
        parser.add_argument('-q', '--quiet', help='set logging to ERROR',
                            action='store_const', dest='loglevel',
                            const=logging.ERROR, default=logging.INFO)
        parser.add_argument('-d', '--debug', help='set logging to DEBUG',
                            action='store_const', dest='loglevel',
                            const=logging.DEBUG, default=logging.INFO)
        parser.add_argument('-v', '--verbose', help='set logging to COMM',
                            action='store_const', dest='loglevel',
                            const=5, default=logging.INFO)
        parser.add_argument('-s', '--settings', help='Settings file module',
                            default=settings)
        parser.sub = parser.add_subparsers()
        # Subcommand Create bot
        parser_createbot = parser.sub.add_parser('createbot', help='Create a new bot directory usign template.')
        parser_createbot.set_defaults(which='createbot')
        parser_createbot.add_argument('name')
        parser_createbot.add_argument('dest', nargs='?', default=None)
        # Subcommand start
        parser_start = parser.sub.add_parser('start', help='Start bot.')
        parser_start.set_defaults(which='start')
        return parser

    def execute(self, parser=None):
        from nekbot.conf import settings
        if parser is None:
            parser = self.parser
        args = parser.parse_args()
        if not args.which in self.no_configuration_required:
            settings.configure(args.settings)
        logging.basicConfig(level=args.loglevel)
        if not hasattr(self, 'command_' + args.which):
            raise ValueError('Comand %s is invalid.' % args.which)
        getattr(self, 'command_' + args.which)(args)

    def command_createbot(self, args):
        if args.dest:
            dest = args.dest
        else:
            dest = args.name
        if not args.dest and os.path.exists(dest):
            sys.stderr.write("Sorry, directory %s exists. I can't create directory.\n" % dest)
            sys.exit(1)
        try:
            copytree(os.path.join(conf_src_dir, 'project_template'), dest)
        except Exception as e:
            sys.stderr.write('Unknown error: %s\n' % e)
        print('Project created as %s' % dest)

    def command_start(self, args):
        from nekbot import NekBot
        nekbot = NekBot()
        nekbot = nekbot.start()
        try:
            nekbot.loop()
        except (KeyboardInterrupt, SystemExit):
            nekbot.close()


PERMISSIONS_TREE = {
    'root': ['admin'],
    'admin': ['execution'],
    'execution': []
}

PROTOCOLS = [
]

PLUGINS = [
    'bot',
    'hello',
]

PERMS_METHODS = [
    'nekbot.core.permissions.perms_from_settings',
]

SYMBOL = '!'


from nekbot.utils.modules import get_module

__author__ = 'nekmo'

from nekbot.conf import settings

def get_permissions_tree(tree):
    def get_permission(perm):
        perms = [perm]
        for subperm in tree[perm]:
            perms += get_permission(subperm)
        return perms
    new_tree = {}
    for perm in tree:
        new_tree[perm] = get_permission(perm)
    return new_tree

permissions_tree = get_permissions_tree(settings.PERMISSIONS_TREE)

def always_true(user, perm):
    """Only for debug!
    """
    return True

def perms_from_settings(user, wanted_perm):
    settings_property = '%s_PERMISSIONS' % user.protocol.name.upper()
    if not hasattr(settings, settings_property):
        return False
    perms = getattr(settings, settings_property)
    if not user.get_id() in perms: return False
    for perm in perms[user.id]:
        if wanted_perm in permissions_tree[perm]:
            return True
    return False

def has_perm(user, perm):
    for perm_method in settings.PERMS_METHODS:
        perm_method = get_module(perm_method)
        if perm_method(user, perm): return True
    return False


# coding=utf-8
from nekbot.core.exceptions import InsufficientPermissions
from nekbot.core.permissions import has_perm

__author__ = 'nekmo'

class Control(object):
    def __init__(self, perm=None, flood=None):
        self.perm = perm

    def check_perms(self, msg):
        if not self.perm: return
        if not has_perm(msg.user, self.perm):
            raise InsufficientPermissions('You need permission %s' % self.perm)

    def check_flood(self, msg):
        pass

    def check(self, msg):
        self.check_perms(msg)
        self.check_flood(msg)

def control(*args, **kwargs):
    """Decorador para añadir restricciones a la ejecución del comando.

    Uso:
        @control(perm='...', flood='...')
        def function():
            pass
    """
    def decorator(f):
        control = Control(*args, **kwargs)
        f.control = control
        return f
    return decorator